/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "domstubs.idl"

[scriptable, uuid(3615a616-571d-4194-bf54-ccf546067b14)]
interface nsIB2GCameraContent : nsISupports
{
    /* temporary solution, waiting for getUserMedia */
    DOMString getCameraURI([optional] in jsval options);
};

[scriptable, uuid(80ad05f8-e5f6-4a36-b25d-5d5a969b365d)]
interface nsIB2GKeyboard : nsISupports
{
    void sendKey(in long aKeyCode, in long aCharCode);
};
